"""
Copyright (C) 2016-2017 Cumulus Networks, inc.

All Rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:.

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/

Cumulus Linux systems: /usr/share/cumulus/EULA.txt


Usage:
    # Linux Basics
    net add hostname <text>
    net show hostname
    net del hostname
    net show version
    net show vrf
    net show bfd [detail] [json]
    net show configuration dhcp [json]

    # interface show commands
    net clear counters
    net show counters [json]
    net show interface (all|bonds|bondmems) [mac|json]
    net show interface [<interface>] [json]
    net show interface <interface> detail
    net show interface pluggables [json]
    net show lldp [<interface>] [json]
    net show system

    # dhcp relay (IPv4)
    net add dhcp relay server <ipv4>
    net del dhcp relay server [<ipv4>]
    net add dhcp relay interface <interface>
    net del dhcp relay interface [<interface>]

    # dhcp relay6 (IPv6)
    net add dhcp relay6 upstream <interface> [<ipv6>]
    net del dhcp relay6 upstream [<interface>]
    net add dhcp relay6 downstream <interface> [<ipv6>]
    net del dhcp relay6 downstream [<interface>]

    # dhcp server (IPv4)
    net add dhcp server <interface>
    net del dhcp server [<interface>]

    # dhcp server6 (IPv6)
    net add dhcp server6 <interface>
    net del dhcp server6 [<interface>]

Options:
    help  : Show this screen and exit
    dhcp  : Dynamic Host Configuration Protocol
    relay : IPv4 relay agent
    relay6 : IPv6 relay agent
    version : A software version number
    server : IPv4 server
    server6 : IPv6 server
    unnumbered : An interface without an IP address
    downstream : A downstream interface
    upstream : An upstream interface
    access     : Summary of physical ports with l2 or l3 config
    bondmems   : Summary of bond members
    bonds      : Summary of bonds
    bridges    : Summary of ports with bridge members
    counters   : net show counters
    interface  : Summary info of all interfaces
    l3         : Summary of ports with an IP
    lldp       : Link Layer Discovery Protocol
    system     : System information
    trunks     : Summary of trunk interfaces
    all        : Everything
    pluggables : SFPs
    json       : Print output in json
    mac        : Media Access Control
"""

from collections import OrderedDict
from nclu import (
    file_set_key_value,
    files_match,
    get_interfaces,
    ifname_expand_glob,
    ifnames_to_glob,
    nclu_check_output,
    NCLU_VERSION,
    netshow2,
    make_pending_diff,
    tabulate_remove_unicode as tru,
    WORKING_DIRECTORY,
    WORKING_DIRECTORIES_PERMISSIONS,
)

from network_docopt import sort_for_humans
from network_docopt.match import ifname_is_glob
from os.path import isdir, isfile
from shutil import copyfile
from subprocess import check_output, STDOUT, CalledProcessError
from tabulate import tabulate
from socket import gethostbyname, gaierror
import json
import logging
import logging.handlers
import os
import re
import shlex
import shutil

log = logging.getLogger("netd")

# File name constants

# DHCP IPv4 relay
ISC_DHCP_RELAY4 = "/etc/default/isc-dhcp-relay"
ISC_DHCP_RELAY_SCRATCHPAD4 = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_relay")
ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_relay.backup")

# DHCP IPv6 relay
ISC_DHCP_RELAY6 = "/etc/default/isc-dhcp-relay6"
ISC_DHCP_RELAY_SCRATCHPAD6 = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_relay6")
ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_relay6.backup")

# DHCP IPv4 server
DHCPD_CONFIG4 = "/etc/dhcp/dhcpd.conf"
DHCPD_PID4 = "/run/dhcpd.pid"
ISC_DHCP_SERVER4 = "/etc/default/isc-dhcp-server"
ISC_DHCP_SERVER_SCRATCHPAD4 = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_server")
ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_server.backup")

# DHCP IPv6 server
DHCPD_CONFIG6 = "/etc/dhcp/dhcpd6.conf"
DHCPD_PID6 = "/run/dhcpd6.pid"
ISC_DHCP_SERVER6 = "/etc/default/isc-dhcp-server6"
ISC_DHCP_SERVER_SCRATCHPAD6 = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_server6")
ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP = os.path.join(WORKING_DIRECTORY, "netmisc", "isc_dhcp_server6.backup")

# hostname
ETC_HOSTS = "/etc/hosts"
ETC_HOSTNAME = "/etc/hostname"
ETC_DHCP_EXIT_HOOKS = "/etc/dhcp/dhclient-exit-hooks.d/dhcp-sethostname"
ETC_HOSTS_SCRATCHPAD = os.path.join(WORKING_DIRECTORY, "netmisc", "etc_hosts")
ETC_HOSTS_SCRATCHPAD_BACKUP = os.path.join(WORKING_DIRECTORY, "netmisc", "etc_hosts.backup")
ETC_HOSTNAME_SCRATCHPAD = os.path.join(WORKING_DIRECTORY, "netmisc", "etc_hostname")
ETC_HOSTNAME_SCRATCHPAD_BACKUP = os.path.join(WORKING_DIRECTORY, "netmisc", "etc_hostname.backup")
ETC_DHCP_EXIT_HOOKS_SCRATCHPAD = os.path.join(WORKING_DIRECTORY, "netmisc", "etc_dhcp_sethostname")
ETC_DHCP_EXIT_HOOKS_SCRATCHPAD_BACKUP = os.path.join(WORKING_DIRECTORY, "netmisc", "etc_dhcp_sethostname.backup")
EXECLIST = os.path.join(WORKING_DIRECTORY, "netmisc", "execlist")


def dhcp_relay_edit(tokens):
    """
    The guts for the following:

    netmisc add dhcp relay server <ipv4>
    netmisc del dhcp relay server [<ipv4>]

    netmisc add dhcp relay interface <interface>
    netmisc del dhcp relay interface [<interface>]

    # Upstream, for IPv6, specifies the interface to which queries from clients, and other relay
    # agents, should be forwarded.  The user has the option to specify a unicast/multicast address.

    netmisc add dhcp relay6 upstream <interface> [<ipv6>]
    netmisc del dhcp relay6 upstream [<interface>]

    # Downstream, for IPv6, specifies the interface on which queries will be received from clients
    # and other relay agents.  The user has the option to specify a link address.

    netmisc add dhcp relay6 downstream <interface> [<ipv6>]
    netmisc del dhcp relay6 downstream [<interface>]

    IPv4 and IPv6 configurations live in different files.
    https://docs.cumulusnetworks.com/display/DOCS/Configuring+DHCP+Relays+and+Servers
    """
    # Does the user want to modify settings for IPv4 (relay) or IPv6 (relay6)?
    RELAY_FILE = ISC_DHCP_RELAY4 if ("relay" in tokens) else ISC_DHCP_RELAY6
    SCRATCHPAD = ISC_DHCP_RELAY_SCRATCHPAD4 if ("relay" in tokens) else ISC_DHCP_RELAY_SCRATCHPAD6

    # Create a default configuration file if a configuration file does not already exist.
    if not isfile(SCRATCHPAD):
        if not isfile(RELAY_FILE):
            with open(RELAY_FILE, 'w') as f:
                f.write('SERVERS=""\n')
                f.write('INTF_CMD=""\n')
                f.write('OPTIONS=""\n')

        copyfile(RELAY_FILE, SCRATCHPAD)

    # Read the configuration file.
    with open(SCRATCHPAD) as f:
        dhcp_config = f.read()

    # Using the appropriate helper function, update the text of dhcp_config.
    # If something goes wrong, the helper functions return None.

    # IPv4 commands
    if "add" in tokens and "relay" in tokens:
        if "server" in tokens:
            # net add dhcp relay server <ipv4>
            dhcp_config = add_dhcp_relay_server(tokens.get("<ipv4>"), dhcp_config)
        elif "interface" in tokens:
            # net add dhcp relay interface <interface>
            dhcp_config = add_dhcp_relay_interface(tokens.get("<interface>"), dhcp_config)

    elif "del" in tokens and "relay" in tokens:
        if "server" in tokens:
            # net del dhcp relay server [<ipv4>]
            dhcp_config = del_dhcp_relay_server(tokens.get("<ipv4>"), dhcp_config)
        elif "interface" in tokens:
            # net del dhcp relay interface [<interface>]
            dhcp_config = del_dhcp_relay_interface(tokens.get("<interface>"), dhcp_config)

    # IPv6 commands
    elif "add" in tokens and "relay6" in tokens:
        if "upstream" in tokens:      # server
            # net add dhcp relay6 upstream <interface> [<ipv6>]
            dhcp_config = add_dhcp_relay6_server(tokens.get("<interface>"), tokens.get("<ipv6>"), dhcp_config)
        elif "downstream" in tokens:  # interface
            # net add dhcp relay6 downstream <interface> [<ipv6>]
            dhcp_config = add_dhcp_relay6_interface(tokens.get("<interface>"), tokens.get("<ipv6>"), dhcp_config)

    elif "del" in tokens and "relay6" in tokens:
        if "upstream" in tokens:      # server
            # net del dhcp relay6 upstream [<interface>]
            dhcp_config = del_dhcp_relay6_server(tokens.get("<interface>"), dhcp_config)
        elif "downstream" in tokens:  # interface
            # net del dhcp relay6 downstream [<interface>]
            dhcp_config = del_dhcp_relay6_interface(tokens.get("<interface>"), dhcp_config)

    else:
        # This shouldn't happen.  It indicates a programming error.
        raise RuntimeError("Cannot determine which DHCP relay command to use.")

    command_ok = dhcp_config is not None

    if command_ok:
        # Write the modified configuration file to disk.
        with open(SCRATCHPAD, 'w') as f:
            f.write(dhcp_config)

    return command_ok


def add_dhcp_relay_server(ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return
    the modified string that results from adding a new server IPv4 address.

    If the given IP is already in the SERVERS string, return None to indicate
    an error.

    net add dhcp relay server <ipv4>
    """
    # Create a string containing the existing server IPs in the configuration file.
    server_line_re = re.compile(r'''^\s*#?\s*SERVERS\s*=\s*"(?P<ips>[.\d ]*)"''', re.IGNORECASE | re.MULTILINE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    server_ips = server_line.group('ips')  # Might be the empty string

    if re.search(r"""\b{0}\b""".format(ip), server_ips) is not None:
        print 'The DHCP relay SERVERS configuration already has the IP address {0}.'.format(ip)
        return None  # Error

    # Add the new IP and/or interface to the configuration file text.
    # Don't include a lead blank if the SERVER string does not already have values.
    if server_ips:
        return server_line_re.sub('SERVERS="{0} {1}"'.format(server_ips, ip), dhcp_config)
    return server_line_re.sub('SERVERS="{0}"'.format(ip), dhcp_config)


def add_dhcp_relay_interface(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from adding one or more new IPv4 relay interfaces.

    If any of the given interfaces are already in the INTF_CMD string, return None to
    indicate an error.

    net add dhcp relay interface <interface>

    interface - A non-empty string; an interface name or glob
    """

    assert isinstance(interface, str) and interface

    # Create a list of strings from the interface(s) the user specified.
    ifnames = ifname_expand_glob(interface)

    # Create a string containing the existing interface commands in the configuration file.
    intf_cmd_line_re = re.compile(r'''^\s*#?\s*INTF_CMD\s*=\s*"(?P<interfaces>[ \S]*?)"''', re.IGNORECASE | re.MULTILINE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMD="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    intf_cmd = intf_cmd_line.group("interfaces")  # A string like "-i swp1 -i swp2"; might be the empty string.

    for ifname in ifnames:
        if re.search(r"""\b{0}\b""".format(ifname), intf_cmd) is not None:
            print "The DHCP relay INTF_CMD configuration already has the interface {0}.".format(ifname)
            return None  # Error

    # Add the new interface(s) to the configuration file text.
    # Don't include a lead blank if the INTF_CMD string does not already have values.
    if intf_cmd:
        return intf_cmd_line_re.sub('INTF_CMD="{0} -i {1}"'.format(intf_cmd, " -i ".join(ifnames)), dhcp_config)
    return intf_cmd_line_re.sub('INTF_CMD="-i {0}"'.format(" -i ".join(ifnames)), dhcp_config)


def del_dhcp_relay_server(ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the modified string that results
    from removing a given server IPv4 address.  If no IPv4 address is given, remove all server IPv4 addresses.

    If the given IP is not in the SERVERS string, return None to indicate an error.

    net del dhcp relay server [<ipv4>]
    """
    server_line_re = re.compile(r'''SERVERS\s*=\s*"(?P<ips>[.\d ]*)"''', re.IGNORECASE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    if not ip:
        # The user did not specifiy an IP to remove.  Remove all IPs.
        return server_line_re.sub('SERVERS=""', dhcp_config)

    # Create a string containing the existing server IPs in the configuration file.
    server_ips = server_line.group("ips")  # Might be the empty string

    if re.search(r"""\b{0}\b""".format(ip), server_ips) is None:
        print 'The DHCP relay SERVERS configuration does not include the IP address {0}.'.format(ip)
        return None  # Error

    # Remove the given IPv4 address from the configuration.
    server_ips = re.sub(ip, '', server_ips)

    # Replace runs of spaces with a single space.
    server_ips = re.sub(r"""\s{2,}""", ' ', server_ips)

    # Replace the contents of SERVERS in the configuration file.
    return server_line_re.sub('SERVERS="{0}"'.format(server_ips.strip()), dhcp_config)


def del_dhcp_relay_interface(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from removing one or more IPv4 relay interfaces.

    If any of the given interfaces are already in the INTF_CMD string, return None to
    indicate an error.

    net del dhcp relay interface [<interface>]
    """
    intf_cmd_line_re = re.compile(r'''INTF_CMD\s*=\s*"(?P<interfaces>[ \S]*?)"''', re.IGNORECASE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    if not interface:
        # The user did not specifiy an interface to remove.  Remove all interfaces.
        return intf_cmd_line_re.sub('INTF_CMD=""', dhcp_config)

    # Create a list of the interface(s) the user specified.  The input might be a glob.
    ifnames = ifname_expand_glob(interface)  # A list of strings

    # Create a string containing the existing interface commands in the configuration file.
    intf_cmd = intf_cmd_line.group('interfaces')  # A string like "-i swp1 -i swp2"; might be the empty string.

    for ifname in ifnames:
        if re.search(r"""\b{0}\b""".format(ifname), intf_cmd) is None:
            print 'The DHCP relay INTF_CMD configuration does not include the interface {0}.'.format(ifname)
            return None  # Error

        # Remove the interface from the configuration.
        intf_cmd = re.sub(r"""-i \b{0}\b""".format(ifname), '', intf_cmd, re.IGNORECASE)

    # Replace runs of spaces with a single space.
    intf_cmd = re.sub(r"""\s{2,}""", ' ', intf_cmd)

    # Replace the contents of INTF_CMD in the configuration file.
    return intf_cmd_line_re.sub(r'''INTF_CMD="{0}"'''.format(intf_cmd.strip()), dhcp_config)


def add_dhcp_relay6_server(interface, ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from adding a new server (upstream) interface name.
    The user optionally can specify an IPv6 address, but this function will accept
    ip=None.

    If the given interface or IPv6 is already in the SERVERS string, return None to
    indicate an error.

    net add dhcp relay6 upstream <interface> [<ipv6>]
    """
    ifnames = ifname_expand_glob(interface) if ifname_is_glob(interface) else [interface]  # A list of strings

    # Create a string containing the existing server interfaces and IPs in the configuration file.
    server_line_re = re.compile(r'''^\s*#?\s*SERVERS\s*=\s*"(?P<interfaces>[-u :\da-z%]*)"''', re.IGNORECASE | re.MULTILINE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    server_interfaces = server_line.group('interfaces')  # Might be the empty string
    server_interfaces_list = server_interfaces.replace('-u', '').strip().split()

    # Test for error conditions.
    for ifname in ifnames:
        if ip:
            new_server_interface = ip + '%' + ifname
        else:
            new_server_interface = ifname

        if new_server_interface in server_interfaces_list:
            print 'The DHCP relay SERVERS configuration already has {0}.'.format(new_server_interface)
            return None  # Error

    # Add the new interface(s), and mabye IPv6, to the configuration file text.
    # Don't include a lead blank if the SERVER string does not already have values.
    if server_interfaces:
        if ip:
            return server_line_re.sub('SERVERS="{0} -u {1}%{2}"'.format(server_interfaces, ip, ' -u {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return server_line_re.sub('SERVERS="{0} -u {1}"'.format(server_interfaces, ' -u '.join(ifnames)), dhcp_config)
    else:
        if ip:
            return server_line_re.sub('SERVERS="-u {0}%{1}"'.format(ip, ' -u {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return server_line_re.sub('SERVERS="-u {0}"'.format(' -u '.join(ifnames)), dhcp_config)


def add_dhcp_relay6_interface(interface, ip, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from adding a new relay interface (downstream).
    The user optionally can specify an IPv6 address, so this function will accept
    ip=None.

    If the given interface or IPv6 is already in the INTF_CMD string, return None to
    indicate an error.

    net add dhcp relay6 downstream <interface> [<ipv6>]
    """

    ifnames = ifname_expand_glob(interface)  # A list of strings

    # Create a string containing the existing server interfaces and IPs in the configuration file.
    intf_cmd_line_re = re.compile(r'''^\s*#?\s*INTF_CMD\s*=\s*"(?P<interfaces>[-l :\da-z%]*)"''', re.IGNORECASE | re.MULTILINE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    intf_cmd = intf_cmd_line.group('interfaces')  # A string like "-l 1::%vlan100 -l vlan101"; might be the empty string.

    # Test for error conditions.
    for ifname in ifnames:
        if re.search(r"""\b{0}\b""".format(ifname), intf_cmd) is not None:
            print 'The DHCP relay INTF_CMD configuration already has the interface {0}.'.format(ifname)
            return None  # Error
    if ip and re.search(r"""\b{0}\b""".format(ip), intf_cmd) is not None:
            print 'The DHCP relay INTF_CMD configuration already has the IPv6 address {0}.'.format(ip)
            return None  # Error

    # Add the new interface(s), and mabye IPv6, to the configuration file text.
    # Don't include a lead blank if the INTF_CMD string does not already have values.
    if intf_cmd:
        if ip:
            return intf_cmd_line_re.sub('INTF_CMD="{0} -l {1}%{2}"'.format(intf_cmd, ip, ' -l {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return intf_cmd_line_re.sub('INTF_CMD="{0} -l {1}"'.format(intf_cmd, ' -l '.join(ifnames)), dhcp_config)
    else:
        if ip:
            return intf_cmd_line_re.sub('INTF_CMD="-l {0}%{1}"'.format(ip, ' -l {0}%'.join(ifnames).format(ip)), dhcp_config)
        else:
            return intf_cmd_line_re.sub('INTF_CMD="-l {0}"'.format(' -l '.join(ifnames)), dhcp_config)


def del_dhcp_relay6_server(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return
    the modified string that results from removing a given server IPv6 interface.
    If no IPv6 interface is given, remove all server IPv6 interfaces.

    If the given IPv6 interface is not in the SERVERS string, return None to indicate
    an error.

    net del dhcp relay6 upstream [<interface>]
    """
    server_line_re = re.compile(r'''SERVERS\s*=\s*"(?P<interfaces>[-u :\da-z%]*)"''', re.IGNORECASE)
    server_line = server_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a SERVER="" line, add one.
    if not server_line:
        dhcp_config += '\nSERVERS=""\n'
        server_line = server_line_re.search(dhcp_config)

    if not interface:
        # The user did not specifiy an IP to remove.  Remove all IPs.
        return server_line_re.sub('SERVERS=""', dhcp_config)

    # Create a list of the interface(s) the user specified.  The input might be a glob.
    ifnames = ifname_expand_glob(interface) if ifname_is_glob(interface) else [interface]  # A list of strings

    # Create a string containing the existing server interfaces in the configuration file.
    server_interfaces = server_line.group('interfaces')  # Might be the empty string

    # Test for an error condition.
    for ifname in ifnames:
        if re.search(r"""\b{0}\b""".format(ifname), server_interfaces) is None:
            print 'The DHCP relay SERVERS configuration does not include the interface {0}.'.format(interface)
            return None  # Error

        # Remove the given IPv6 interface from the configuration.
        server_interfaces = re.sub('-u (?:[:\da-f]+%)?{0}'.format(ifname), '', server_interfaces)

    # Replace runs of spaces with a single space.
    server_interfaces = re.sub(r"""\s{2,}""", ' ', server_interfaces)

    # Replace the contents of SERVERS in the configuration file.
    return server_line_re.sub('SERVERS="{0}"'.format(server_interfaces.strip()), dhcp_config)


def del_dhcp_relay6_interface(interface, dhcp_config):
    """
    Given a string containing the contents of the configuration file, return the
    modified string that results from removing one or more IPv6 relay interfaces.

    If any of the given interfaces are already in the INTF_CMD string, return None to
    indicate an error.

    net del dhcp relay6 downstream [<interface>]
    """

    intf_cmd_line_re = re.compile(r'''INTF_CMD\s*=\s*"(?P<interfaces>[-l :\da-z%]*)"''', re.IGNORECASE)
    intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not intf_cmd_line:
        dhcp_config += '\nINTF_CMD=""\n'
        intf_cmd_line = intf_cmd_line_re.search(dhcp_config)

    if not interface:
        # The user did not specifiy an interface to remove.  Remove all interfaces.
        return intf_cmd_line_re.sub('INTF_CMD=""', dhcp_config)

    # Create a list of the interface(s) the user specified.  The input might be a glob.
    ifnames = ifname_expand_glob(interface)  # A list of strings

    # Create a string containing the existing interface commands in the configuration file.
    intf_cmd = intf_cmd_line.group('interfaces')  # A string like "-l 1::%vlan100"; might be the empty string.

    for ifname in ifnames:
        if re.search(r"""\b{0}\b""".format(ifname), intf_cmd) is None:
            print 'The DHCP relay INTF_CMD configuration does not include the interface {0}.'.format(ifname)
            return None  # Error

        # Remove the interface from the configuration.
        intf_cmd = re.sub(r"""-l \b(?:[:\da-f]+%)?{0}\b""".format(ifname), '', intf_cmd, re.IGNORECASE)

    # Replace runs of spaces with a single space.
    intf_cmd = re.sub(r"""\s{2,}""", ' ', intf_cmd)

    # Replace the contents of INTF_CMD in the configuration file.
    return intf_cmd_line_re.sub(r'''INTF_CMD="{0}"'''.format(intf_cmd.strip()), dhcp_config)


def dhcp_server_edit(tokens):
    """
    The guts for the following:

    netmisc add dhcp server <interface>
    netmisc del dhcp server [<interface>]
    netmisc add dhcp server6 <interface>
    netmisc del dhcp server6 [<interface>]

    IPv4 and IPv6 configurations live in different files.
    https://docs.cumulusnetworks.com/display/DOCS/Configuring+DHCP+Relays+and+Servers
    """

    command_ok = True

    # Does the user want to modify settings for IPv4 (server) or IPv6 (server6)?
    SERVER_FILE = ISC_DHCP_SERVER4 if ("server" in tokens) else ISC_DHCP_SERVER6
    SCRATCHPAD = ISC_DHCP_SERVER_SCRATCHPAD4 if ("server" in tokens) else ISC_DHCP_SERVER_SCRATCHPAD6
    DHCPD_CONFIG = DHCPD_CONFIG4 if ("server" in tokens) else DHCPD_CONFIG6
    DHCPD_PID = DHCPD_PID4 if ("server" in tokens) else DHCPD_PID6

    # Create a default configuration file if a configuration file does not already exist.
    if not isfile(SCRATCHPAD):
        if not isfile(SERVER_FILE):
            with open(SERVER_FILE, 'w') as f:
                f.write('DHCPD_CONF="-cf {0}"\n'.format(DHCPD_CONFIG))
                f.write('DHCPD_PID="-pf {0}"\n'.format(DHCPD_PID))
                f.write('OPTIONS=""\n')
                f.write('INTERFACES=""\n')

        copyfile(SERVER_FILE, SCRATCHPAD)

    # Read the configuration file.
    with open(SCRATCHPAD) as f:
        dhcp_config = f.read()

    # Get the interface(s) the user specified, if any.
    ifname = tokens.get("<interface>")
    if ifname is not None:  # Might be None for a "del" command
        ifnames = ifname_expand_glob(ifname) if ifname_is_glob(ifname) else [ifname]  # A list of strings

    # Create a string containing the existing interface commands in the configuration file.
    interfaces_re = re.compile(r'''^\s*#?\s*INTERFACES\s*=\s*"(?P<interfaces>[ \S]*?)"''', re.IGNORECASE | re.MULTILINE)
    interfaces_obj = interfaces_re.search(dhcp_config)

    # If our baseline DHCP config file does not contain a INTF_CMDS="" line, add one.
    if not interfaces_obj:
        dhcp_config += '\nINTERFACES=""\n'
        interfaces_obj = interfaces_re.search(dhcp_config)

    interfaces = interfaces_obj.group('interfaces')  # A string like "swp1 swp2"; might be the empty string.

    if "add" in tokens:     # net add dhcp server <interface>
        interfaces_to_add = []

        for ifname in ifnames:
            if ifname in interfaces:
                print 'DHCP server INTERFACES configuration already has {0}.'.format(ifname)
                command_ok = False
            else:
                interfaces_to_add.append(ifname)

        if interfaces_to_add:
            # Add the interface(s) to the configuration file text.  Don't include a lead blank if
            # the INTERFACES string does not already have values.
            if interfaces:
                dhcp_config = interfaces_re.sub('INTERFACES="{0} {1}"'.format(interfaces, ' '.join(interfaces_to_add)), dhcp_config)
            else:
                dhcp_config = interfaces_re.sub('INTERFACES="{0}"'.format(' '.join(interfaces_to_add)), dhcp_config)

    elif "del" in tokens:  # net del dhcp server ...
        if ifname is not None:
            # net del dhcp server <interface>
            # Remove the given interface from the configuration file text.
            for ifname in ifnames:
                if ifname in interfaces:
                    interfaces = re.sub(ifname, '', interfaces, re.IGNORECASE)
                else:
                    print 'DHCP server INTERFACES configuration does not have {0}.'.format(ifname)
                    command_ok = False

            # Replace multi-spaces in the interfaces string with single spaces.
            interfaces = re.sub(r"""\s{2,}""", ' ', interfaces)

            # Update the INTERFACES line in the configuration file text.
            dhcp_config = interfaces_re.sub('INTERFACES="{0}"'.format(interfaces.strip()), dhcp_config)

        else:
            # net del dhcp server
            # The user did not specifiy an interface to remove.  Remove all interfaces.
            dhcp_config = interfaces_re.sub('INTERFACES=""', dhcp_config)

    if command_ok:
        # Write the modified configuration file to disk.
        with open(SCRATCHPAD, 'w') as f:
            f.write(dhcp_config)

    return command_ok


def hostname_valid(hostname):
    return re.match(r"""[-a-zA-Z\d\.]{1,63}""", hostname) is not None


def is_fqdn(name):
    return '.' in name


def hostname_edit(new_hostname):
    if not hostname_valid(new_hostname):
        print "\"{0}\" is an invalid hostname.  Hostnames may only contain letters, numbers, or '-'.".format(new_hostname)
        return False

    if is_fqdn(new_hostname):
        try:
            # Check with DNS to see if this (potential) FQDN can be resolved.
            gethostbyname(new_hostname)
        except gaierror:
            # The FQDN cannot be resolved. Warn the user.
            print "WARNING: DNS could not resolve the hostname entered. Please make sure the Fully Qualified Domain Name is valid."

    # Hostname changes should be made in /etc/hostname and /etc/hosts.
    # Execute the 'hostname' command to persist the change.

    # execlist
    with open(EXECLIST, 'w') as f:
        f.write("hostname {0}\n".format(new_hostname))
        f.write('echo "You must re-login for your bash prompt to reflect the hostname change"\n')

    # /etc/hostname
    with open(ETC_HOSTNAME_SCRATCHPAD, 'w') as f:
        f.write(new_hostname + '\n')

    try:
        old_hostname = check_output(["/bin/hostname"]).strip()
    except CalledProcessError as e:
        print str(e)
        return False

    with open(ETC_HOSTS) as f:
        etc_hosts_lines = [line.strip() for line in f]
        old_hostname_re = re.compile(r"""\b{0}\b""".format(old_hostname), re.IGNORECASE)

    with open(ETC_HOSTS_SCRATCHPAD, 'w') as f:
        for line in etc_hosts_lines:
            if not line.startswith('#'):
                line = old_hostname_re.sub(new_hostname, line)
            f.write(line + '\n')

    # The action succeeded.
    return True


def check_dhcp_sethostname_yes():
    """
    Checks for SETHOSTNAME="yes"
    Returns True if set to yes,
    Returns False otherwise.
    """
    if isfile(ETC_DHCP_EXIT_HOOKS):
        with open(ETC_DHCP_EXIT_HOOKS) as f:
            for line in f:
                if line.startswith('SETHOSTNAME="yes"'):
                    return True
    return False


def modify_dhcp_exit_hooks(old, new):
    """
    Modify the DHCP exit hooks sethostname file.
    If old is "yes" and new is "no", SETHOSTNAME="yes" will change to SETHOSTNAME="no"

    If NCLU sets the hostname, the DHCP sethostname file should not let DHCP change
    the hostname on system reboot.
    """

    with open(ETC_DHCP_EXIT_HOOKS) as f:
        exit_hooks_l = []
        for line in f:
            if line.startswith('SETHOSTNAME="{}"'.format(old)):
                line = 'SETHOSTNAME="{}"\n'.format(new)
            exit_hooks_l.append(line)
        new_exit_hooks = ''.join(exit_hooks_l)

    with open(ETC_DHCP_EXIT_HOOKS_SCRATCHPAD, 'w') as f:
        f.write(new_exit_hooks)


def get_pluggables(ifnames):
    data = OrderedDict()

    for ifname in ifnames:
        if re.match('swp\d+', ifname) or re.match('swp\d+s\d+', ifname):
            try:
                output = check_output(['/sbin/ethtool', '-m', ifname], stderr=STDOUT)
            except Exception:
                continue

            data[ifname] = {}

            for line in output.splitlines():
                # compress multiple spaces
                line = ' '.join(line.strip().split())

                if 'No such device' in line or 'Cannot get Module EEPROM data' in line:
                    break

                (key, value) = line.split(':', 1)
                key = key.strip()
                value = value.strip()

                if value:
                    if key == 'Identifier':
                        (number, description) = value.split(' ', 1)

                        if number.startswith('0x'):
                            number = int(number, 16)
                        elif number.isdigit():
                            number = int(number)

                        if 'no module' in description:
                            description = '(empty)'

                        data[ifname]['identifier'] = number
                        data[ifname]['identifierDescription'] = description
                    elif key == 'Vendor PN':
                        data[ifname]['vendorPartNumber'] = value
                    elif key == 'Vendor rev':
                        data[ifname]['vendorRev'] = value
                    elif key == 'Vendor name':
                        data[ifname]['vendorName'] = value
                    elif key == 'Vendor SN':
                        data[ifname]['vendorSerialNumber'] = value

    return data


def show_pluggables(cli):

    if cli.args.get("<interface>") is not None:
        assert isinstance(cli.args.get("<interface>"), str), "This should not be a list."
        ifnames = [cli.args.get("<interface>")]
    else:
        ifnames = get_interfaces()

    data = get_pluggables(ifnames)

    if cli.args.get("json") is not None:
        reply = json.dumps(data, indent=4)
    else:
        header = ("Interface", "Identifier", "Vendor Name", "Vendor PN", "Vendor SN", "Vendor Rev")
        table = []
        for (ifname, value) in data.iteritems():
            table.append((ifname,
                          "0x%02x %s" % (value.get("identifier", 0), value.get("identifierDescription")),
                          value.get("vendorName"),
                          value.get("vendorPartNumber"),
                          value.get("vendorSerialNumber"),
                          value.get("vendorRev")))
        reply = '\n' + tabulate(tru(table), header, tablefmt="simple") + "\n\n"

    return reply


def get_ifname_to_master():
    """
    Example output:

54: swp52: <BROADCAST,MULTICAST> mtu 1500 qdisc pfifo_fast state DOWN mode DEFAULT group default qlen 1000\    link/ether 00:e0:ec:27:68:e6 brd ff:ff:ff:ff:ff:ff
55: swp53: <BROADCAST,MULTICAST> mtu 1500 qdisc pfifo_fast state DOWN mode DEFAULT group default qlen 1000\    link/ether 00:e0:ec:27:68:ea brd ff:ff:ff:ff:ff:ff
56: swp54: <BROADCAST,MULTICAST> mtu 1500 qdisc pfifo_fast state DOWN mode DEFAULT group default qlen 1000\    link/ether 00:e0:ec:27:68:ee brd ff:ff:ff:ff:ff:ff
370: vrf1001: <NOARP,MASTER,UP,LOWER_UP> mtu 65536 qdisc pfifo_fast state UP mode DEFAULT group default qlen 1000\    link/ether 4a:2d:c6:02:a9:71 brd ff:ff:ff:ff:ff:ff
371: swp4.1001@swp4: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc noqueue master vrf1001 state UP mode DEFAULT group default \    link/ether 00:e0:ec:27:68:ad brd ff:ff:ff:ff:ff:ff
372: swp3.1001@swp3: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc noqueue master vrf1001 state UP mode DEFAULT group default \    link/ether 00:e0:ec:27:68:ac brd ff:ff:ff:ff:ff:ff
    """
    result = {}
    output = check_output(['/bin/ip', '-o', 'link', 'show'])

    for line in output.splitlines():
        re_master = re.search('^\d+:\s+(\S+):.*master\s+(\S+)', line)

        if re_master:
            ifname = re_master.group(1)

            # For swp3.1057@swp3: keep the "swp3.1057'
            if '@' in ifname:
                ifname = ifname.split('@')[0]

            result[ifname] = re_master.group(2)

    return result


def get_bgp_peers():
    """
    Return a set of all bgp peers
    """
    try:
        data = json.loads(check_output(['/usr/bin/vtysh', '-c', 'show ip bgp neighbor json']))
        return set(data.keys())
    except Exception:
        return set()


class ConfigWrapper(object):
    """
    Note that this class does not inherit from __init__.py:ConfigWrapperBase.
    """

    def __init__(self, show_linux_command, color_diffs):
        self.show_linux_command = show_linux_command
        self.color_diffs = color_diffs

    def get_managed_files(self):
        """
        TODO - This could be a free function, but the "wrappers" execution loop for rollbacks
        expects it to be a method of a ConfigWrapper instance.
        """

        return [ISC_DHCP_RELAY4, ISC_DHCP_RELAY6, ISC_DHCP_SERVER4, ISC_DHCP_SERVER6, ETC_HOSTNAME, ETC_HOSTS, ETC_DHCP_EXIT_HOOKS]

    def get_service_cmd_for_file(self, filename):
        """
        TODO - This could be a free function, but the "wrappers" execution loop for rollbacks
        expects it to be a method of a ConfigWrapper instance.
        """

        cmds = []

        if filename == ISC_DHCP_RELAY4:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcrelay.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcrelay.service'))

        elif filename == ISC_DHCP_RELAY6:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcrelay6.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcrelay6.service'))

        elif filename == ISC_DHCP_SERVER4:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcpd.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcpd.service'))

        elif filename == ISC_DHCP_SERVER6:
            cmds.append(('/bin/systemctl', 'reset-failed', 'dhcpd6.service'))
            cmds.append(('/bin/systemctl', 'restart', 'dhcpd6.service'))

        elif filename == ETC_HOSTNAME:
            new_hostname = None

            with open(ETC_HOSTNAME) as f:
                for line in f:
                    line = line.strip()

                    if line:
                        new_hostname = line
                        break

            if new_hostname:
                cmds.append(('hostname', new_hostname))
                cmds.append(('hostnamectl', 'set-hostname', new_hostname, '--static'))

        return cmds

    def abort_pending(self):
        """
        TODO - This could be a free function, but the "wrappers" execution loop
        expects it to be a method of a ConfigWrapper instance.
        """

        if isdir(os.path.join(WORKING_DIRECTORY, "netmisc")):
            for filename in os.listdir(os.path.join(WORKING_DIRECTORY, "netmisc")):
                full_filename = os.path.join(WORKING_DIRECTORY, "netmisc", filename)

                if os.path.isdir(full_filename):
                    shutil.rmtree(full_filename)
                else:
                    os.remove(full_filename)

    def del_all(self, cli):
        """
        Remove all DHCP configuration file INTF_CMD, INTERFACES, OPTIONS, and SERVERS entries, if any.
        If any of the configuration files are missing, the call to file_set_key_value should create
        scratchpad files with defaults.

        TODO - This could be a free function, but the "wrappers" execution loop
        expects it to be a method of a ConfigWrapper instance.
        """

        # TODO - Move all this DHCP code into its own wrapper.

        # "copyfile", used below, raises OSError is the destination directory does not exist.
        if not isdir(os.path.join(WORKING_DIRECTORY, "netmisc")):
            os.makedirs(os.path.join(WORKING_DIRECTORY, "netmisc"))

        # WARNING - The deleted values must be set to an empty string with double quotes because regular
        # expressions elsewhere in this file, in code related to add|del, specifically look for double quotes.
        # Setting a key to nothing (i.e. SERVERS=<blank>) will break things.

        if isfile(ISC_DHCP_RELAY4) or isfile(ISC_DHCP_RELAY_SCRATCHPAD4):
            if isfile(ISC_DHCP_RELAY4) and not isfile(ISC_DHCP_RELAY_SCRATCHPAD4):
                copyfile(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD4, "SERVERS", '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD4, "INTF_CMD", '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD4, "OPTIONS", '""', True)

        if isfile(ISC_DHCP_RELAY6) or isfile(ISC_DHCP_RELAY_SCRATCHPAD6):
            if isfile(ISC_DHCP_RELAY6) and not isfile(ISC_DHCP_RELAY_SCRATCHPAD6):
                copyfile(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD6, "SERVERS", '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD6, "INTF_CMD", '""', True)
            file_set_key_value(ISC_DHCP_RELAY_SCRATCHPAD6, "OPTIONS", '""', True)

        if isfile(ISC_DHCP_SERVER4) or isfile(ISC_DHCP_SERVER_SCRATCHPAD4):
            if isfile(ISC_DHCP_SERVER4) and not isfile(ISC_DHCP_SERVER_SCRATCHPAD4):
                copyfile(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD4, "INTERFACES", '""', True)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD4, "OPTIONS", '""', True)

        if isfile(ISC_DHCP_SERVER6) or isfile(ISC_DHCP_SERVER_SCRATCHPAD6):
            if isfile(ISC_DHCP_SERVER6) and not isfile(ISC_DHCP_SERVER_SCRATCHPAD6):
                copyfile(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD6, "INTERFACES", '""', True)
            file_set_key_value(ISC_DHCP_SERVER_SCRATCHPAD6, "OPTIONS", '""', True)

    def commit_pending(self, verbose):
        """
        Persistent changes, and return a 3-tuple of (bool, str, [str]).  The boolean indicates
        if the commit succeeded.  The string includes information, including errors, for the user.
        The list contains the names of the files modified.
        """

        # TODO - Move all this DHCP code into its own plug-in.

        # These files already shouldn't exist, but better safe than sorry.
        for backup_file in (ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP,
                            ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP,
                            ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP,
                            ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP,
                            ETC_HOSTS_SCRATCHPAD_BACKUP,
                            ETC_HOSTNAME_SCRATCHPAD_BACKUP,
                            ETC_DHCP_EXIT_HOOKS_SCRATCHPAD_BACKUP):
            if isfile(backup_file):
                os.remove(backup_file)

        commit_ok = True
        message = ''
        files_modified = []

        # "copyfile", used below, raises OSError is the destination directory does not exist.
        if not isdir(os.path.join(WORKING_DIRECTORY, "netmisc")):
            os.makedirs(os.path.join(WORKING_DIRECTORY, "netmisc"))

        # DHCP IPv4 relay
        if isfile(ISC_DHCP_RELAY_SCRATCHPAD4) and not files_match(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4):

            with open(ISC_DHCP_RELAY_SCRATCHPAD4) as f:
                isc_dhcp_relay_scratchpad4_contents = f.read()

            # Don't permit configuring a DHCP relay interface without having a configured relay server.
            if re.search(r'''INTF_CMD\s*=\s*"\s*\S.*?"''', isc_dhcp_relay_scratchpad4_contents) is not None and \
                    re.search(r'''SERVER\s*=\s*"\s*"''', isc_dhcp_relay_scratchpad4_contents) is not None:
                return (False, "First run \"net add dhcp relay server <ipv4>\".", [])

            shutil.copyfile(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP)
            shutil.copyfile(ISC_DHCP_RELAY_SCRATCHPAD4, ISC_DHCP_RELAY4)
            files_modified.append(ISC_DHCP_RELAY4)

            # Service resets
            for command in (["/bin/systemctl", "enable", "dhcrelay.service"],
                            ["/bin/systemctl", "reset-failed", "dhcrelay.service"],
                            ["/bin/systemctl", "restart", "dhcrelay.service"]):

                if self.show_linux_command:
                    message += "exec: {0}\n".format(' '.join(command))

                try:
                    output = check_output(command, stderr=STDOUT)
                    if verbose:
                        message += output
                except CalledProcessError as e:
                    commit_ok = False
                    message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)
                    revert_dhcp(True, True)
                    return (False, message, [])

        # DHCP IPv4 server
        if commit_ok and isfile(ISC_DHCP_SERVER_SCRATCHPAD4) and not files_match(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4):

            shutil.copyfile(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP)
            shutil.copyfile(ISC_DHCP_SERVER_SCRATCHPAD4, ISC_DHCP_SERVER4)
            files_modified.append(ISC_DHCP_SERVER4)

            # Service resets
            for command in (["/bin/systemctl", "enable", "dhcpd.service"],
                            ["/bin/systemctl", "reset-failed", "dhcpd.service"],
                            ["/bin/systemctl", "restart", "dhcpd.service"]):

                if self.show_linux_command:
                    message += "exec: {0]".format(' '.join(command))

                try:
                    output = check_output(command, stderr=STDOUT)
                    if verbose:
                        message += output
                except CalledProcessError as e:
                    commit_ok = False
                    message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)

                    try:
                        output = check_output(["/bin/systemctl", "status", "dhcpd.service"], stderr=STDOUT)
                        message += '\n/bin/systemctl status dhcpd.service'
                        message += '======================================'
                        message += output
                    except CalledProcessError as e:
                        message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)
                        revert_dhcp(True, True)
                        revert_dhcp(True, False)
                        return (False, message, [])

        # DHCP IPv6 relay
        if commit_ok and isfile(ISC_DHCP_RELAY_SCRATCHPAD6) and not files_match(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6):

            with open(ISC_DHCP_RELAY_SCRATCHPAD6) as f:
                isc_dhcp_relay_scratchpad6_contents = f.read()

            # Don't permit configuring a DHCP relay interface without having a configured relay server.
            if re.search(r'''INTF_CMD\s*=\s*"\s*\S.*?"''', isc_dhcp_relay_scratchpad6_contents) is not None and \
                    re.search(r'''SERVER\s*=\s*"\s*"''', isc_dhcp_relay_scratchpad6_contents) is not None:

                commit_ok = False
                message += "First run \"net add dhcp relay6 downstream <interface> [<ipv6>]\"."
                revert_dhcp(True, True)
                revert_dhcp(True, False)
                return (False, message, [])

            else:
                shutil.copyfile(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP)
                shutil.copyfile(ISC_DHCP_RELAY_SCRATCHPAD6, ISC_DHCP_RELAY6)
                files_modified.append(ISC_DHCP_RELAY6)

                # Service resets
                for command in (["/bin/systemctl", "enable", "dhcrelay6.service"],
                                ["/bin/systemctl", "reset-failed", "dhcrelay6.service"],
                                ["/bin/systemctl", "restart", "dhcrelay6.service"]):

                    if self.show_linux_command:
                        message += "exec: {0}".format(' '.join(command))

                    try:
                        output = check_output(command, stderr=STDOUT)
                        if verbose:
                            message += output
                    except CalledProcessError as e:
                        commit_ok = False
                        message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)

                        try:
                            output = check_output(['/bin/systemctl', 'status', 'dhcrelay6.service'], stderr=STDOUT)
                            message += '\n/bin/systemctl status dhcrelay6.service'
                            message += '======================================'
                            message += output
                        except CalledProcessError as e:
                            message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)
                            revert_dhcp(True, True)
                            revert_dhcp(True, False)
                            revert_dhcp(False, True)
                            return (False, message, [])

        # DHCP IPv6 server
        if commit_ok and isfile(ISC_DHCP_SERVER_SCRATCHPAD6) and not files_match(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6):

            shutil.copyfile(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP)
            shutil.copyfile(ISC_DHCP_SERVER_SCRATCHPAD6, ISC_DHCP_SERVER6)
            files_modified.append(ISC_DHCP_SERVER6)

            # Service resets
            for command in (["/bin/systemctl", "enable", "dhcpd6.service"],
                            ["/bin/systemctl", "reset-failed", "dhcpd6.service"],
                            ["/bin/systemctl", "restart", "dhcpd6.service"]):

                if self.show_linux_command:
                    message += "exec: {0]".format(' '.join(command))

                try:
                    output = check_output(command, stderr=STDOUT)
                    if verbose:
                        message += output
                except CalledProcessError as e:
                    commit_ok = False
                    message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)

                    try:
                        output = check_output(['/bin/systemctl', 'status', 'dhcpd6.service'], stderr=STDOUT)
                        message += '\n/bin/systemctl status dhcpd6.service'
                        message += '======================================'
                        message += output
                    except CalledProcessError as e:
                        message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)
                        revert_dhcp(True, True)
                        revert_dhcp(True, False)
                        revert_dhcp(False, True)
                        revert_dhcp(False, False)
                        return (False, message, [])

        # hostname
        if commit_ok and isfile(ETC_HOSTS_SCRATCHPAD) and not files_match(ETC_HOSTS, ETC_HOSTS_SCRATCHPAD):
            shutil.copyfile(ETC_HOSTS, ETC_HOSTS_SCRATCHPAD_BACKUP)
            shutil.copyfile(ETC_HOSTS_SCRATCHPAD, ETC_HOSTS)
            files_modified.append(ETC_HOSTS)

        if commit_ok and isfile(ETC_HOSTNAME_SCRATCHPAD) and not files_match(ETC_HOSTNAME, ETC_HOSTNAME_SCRATCHPAD):
            shutil.copyfile(ETC_HOSTNAME, ETC_HOSTNAME_SCRATCHPAD_BACKUP)
            shutil.copyfile(ETC_HOSTNAME_SCRATCHPAD, ETC_HOSTNAME)
            files_modified.append(ETC_HOSTNAME)

        if commit_ok and isfile(ETC_DHCP_EXIT_HOOKS_SCRATCHPAD) and not files_match(ETC_DHCP_EXIT_HOOKS, ETC_DHCP_EXIT_HOOKS_SCRATCHPAD):
            shutil.copyfile(ETC_DHCP_EXIT_HOOKS, ETC_DHCP_EXIT_HOOKS_SCRATCHPAD_BACKUP)
            shutil.copyfile(ETC_DHCP_EXIT_HOOKS_SCRATCHPAD, ETC_DHCP_EXIT_HOOKS)
            files_modified.append(ETC_DHCP_EXIT_HOOKS)

        if commit_ok and isfile(EXECLIST):
            with open(EXECLIST) as f:
                for line in f.readlines():
                    command = shlex.split(line.strip())

                    if self.show_linux_command:
                        message += "exec: %s".format(' '.join(command))

                    try:
                        output = check_output(command, stderr=STDOUT)
                        if verbose:
                            message += output
                    except CalledProcessError as e:
                        commit_ok = False
                        message += "\n\"{0}\" failed:\n{1}\n".format(' '.join(command), e)
                        revert_dhcp(True, True)
                        revert_dhcp(True, False)
                        revert_dhcp(False, True)
                        revert_dhcp(False, False)
                        revert_hostname()
                        return (False, message, [])

            os.remove(EXECLIST)

        # The commit succeeded.  Remove the scratchpad and backup files.
        for filename in (ISC_DHCP_RELAY_SCRATCHPAD4,
                         ISC_DHCP_SERVER_SCRATCHPAD4,
                         ISC_DHCP_RELAY_SCRATCHPAD6,
                         ISC_DHCP_SERVER_SCRATCHPAD6,
                         ETC_HOSTS_SCRATCHPAD,
                         ETC_HOSTNAME_SCRATCHPAD,
                         ETC_DHCP_EXIT_HOOKS_SCRATCHPAD,
                         ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP,
                         ETC_HOSTS_SCRATCHPAD_BACKUP,
                         ETC_HOSTNAME_SCRATCHPAD_BACKUP):
            if isfile(filename):
                os.remove(filename)

        return (commit_ok, message, files_modified)

    def get_pending(self):
        """
        Return the pending diff as a string.  If there are no changes, return the empty string.
        """

        the_diff = make_pending_diff(ISC_DHCP_RELAY4, ISC_DHCP_RELAY_SCRATCHPAD4, self.color_diffs) \
                 + make_pending_diff(ISC_DHCP_RELAY6, ISC_DHCP_RELAY_SCRATCHPAD6, self.color_diffs) \
                 + make_pending_diff(ISC_DHCP_SERVER4, ISC_DHCP_SERVER_SCRATCHPAD4, self.color_diffs) \
                 + make_pending_diff(ISC_DHCP_SERVER6, ISC_DHCP_SERVER_SCRATCHPAD6, self.color_diffs) \
                 + make_pending_diff(ETC_HOSTS, ETC_HOSTS_SCRATCHPAD, self.color_diffs) \
                 + make_pending_diff(ETC_DHCP_EXIT_HOOKS, ETC_DHCP_EXIT_HOOKS_SCRATCHPAD, self.color_diffs) \
                 + make_pending_diff(ETC_HOSTNAME, ETC_HOSTNAME_SCRATCHPAD, self.color_diffs)

        return the_diff

    def show_config_files(self, user_may_edit):
        # TODO - This method doesn't need user_may_edit.

        reply = ''

        for filename in (ETC_HOSTNAME,
                         ETC_HOSTS,
                         ETC_DHCP_EXIT_HOOKS,
                         ISC_DHCP_RELAY4,
                         ISC_DHCP_RELAY6,
                         ISC_DHCP_SERVER4,
                         ISC_DHCP_SERVER6):
            if isfile(filename) and os.path.getsize(filename) > 0:
                reply += filename + '\n'
                reply += '=' * len(filename) + '\n'

                with open(filename) as f:
                    reply += f.read() + '\n'

        return reply

    def show_config_commands(self, user_may_edit, ifupdown2_wrapper):
        commands = []
        hostname = get_configured_hostname()

        if not check_dhcp_sethostname_yes() and hostname is not None:
            commands.append("net add hostname {0}".format(hostname))

        (dhcp_relay_servers, dhcp_relay_interfaces, dhcp_servers, dhcp_relay6_servers, dhcp_relay6_interfaces, dhcp_servers6) = get_dhcp_configuration()

        # DHCP IPv4
        commands.extend(["net add dhcp relay server {0}".format(server) for server in dhcp_relay_servers])
        commands.extend(["net add dhcp relay interface {0}".format(interface) for interface in dhcp_relay_interfaces])
        commands.extend(["net add dhcp server {0}".format(server) for server in dhcp_servers])

        # DHCP IPv6
        commands.extend(["net add dhcp relay6 server {0}".format(server) for server in dhcp_relay6_servers])
        commands.extend(["net add dhcp relay6 interface {0}".format(interface) for interface in dhcp_relay6_interfaces])
        commands.extend(["net add dhcp server6 {0}".format(server) for server in dhcp_servers6])

        return (commands, [])

    def show_config_summary(self, user_may_edit, summary, print_json=False):
        """
        net show configuration [dhcp [json]]

        Populate the "summary" dictionary with DHCP information.  If the configuration for
        DHCP was not specifically requested, also provide information about the hostname.

        summary - A dictionary to hold results (call by reference)

        TODO - This should just return a value and not require passing a dictionary to hold results.
        """

        assert isinstance(summary, dict)

        hostname = get_configured_hostname()
        if not check_dhcp_sethostname_yes() and hostname is not None:
            # net show configuration [json]
            summary["hostname %s" % hostname] = []

        (dhcp_relay_servers, dhcp_relay_interfaces, dhcp_servers, dhcp_relay6_servers, dhcp_relay6_interfaces, dhcp_servers6) = get_dhcp_configuration()

        # IPv4 relay
        if dhcp_relay_servers or dhcp_relay_interfaces:
            summary["dhcp relay"] = OrderedDict()

            if dhcp_relay_servers:
                summary["dhcp relay"]["SERVERS"] = dhcp_relay_servers

            if dhcp_relay_interfaces:
                summary["dhcp relay"]["INTF_CMD"] = dhcp_relay_interfaces

        # IPv4 server
        if dhcp_servers:
            summary["dhcp server"] = OrderedDict()
            summary["dhcp server"]["INTERFACES"] = dhcp_servers

        # IPv6 relay
        if dhcp_relay6_servers or dhcp_relay6_interfaces:
            summary["dhcp relay6"] = OrderedDict()

            if dhcp_relay6_servers:
                summary["dhcp relay6"]["SERVERS"] = dhcp_relay6_servers

            if dhcp_relay6_interfaces:
                summary["dhcp relay6"]["INTF_CMD"] = dhcp_relay6_interfaces

        # IPv6 server
        if dhcp_servers6:
            summary["dhcp server6"] = OrderedDict()
            summary["dhcp server6"]["INTERFACES"] = dhcp_servers6

        if print_json:
            print json.dumps(summary, indent=4)

    def eval_command_line_args(self, cli):
        """
        NetworkDocopt will populate the cli.args dictionary based on the command line args entered by the user.

        TODO - This could be a free function, but the "wrappers" execution loop in execute_command.py
        expects it to be a method of a ConfigWrapper instance.
        """

        tokens = cli.args
        command_ok = None
        output = ''

        if not os.path.isdir(os.path.join(WORKING_DIRECTORY, "netmisc")):
            # Create the scratchpad directory.
            os.makedirs(os.path.join(WORKING_DIRECTORY, "netmisc"), mode=WORKING_DIRECTORIES_PERMISSIONS)

        if "clear" in tokens and "counters" in tokens:
            # net clear counters

            cmd = ["/usr/cumulus/bin/cl-netstat", "-c"]
            try:
                check_output(cmd, stderr=STDOUT)
                command_ok = True
            except Exception as e:
                output = "\"{0}\" failed.\n{1}".format(' '.join(cmd), e)
                command_ok = False

        elif "dhcp" in tokens:

            if "relay" in tokens or "relay6" in tokens:
                # net add|del dhcp relay|relay6 ...
                command_ok = dhcp_relay_edit(tokens)
            elif "server" in tokens or "server6" in tokens:
                # net add|del dhcp server|server6 ...
                command_ok = dhcp_server_edit(tokens)

        elif "hostname" in tokens:
            # net add|del hostname <text>
            if "add" in tokens:
                command_ok = hostname_edit(tokens.get("<text>"))
                modify_dhcp_exit_hooks("yes", "no")
            elif "del" in tokens:
                command_ok = hostname_edit("cumulus")
                modify_dhcp_exit_hooks("no", "yes")
            else:
                # net show hostname
                try:
                    output = check_output("/bin/hostname")
                    command_ok = True
                except CalledProcessError as e:
                    output = "Could not get the hostname: {0}".format(e.output)
                    command_ok = False

        elif "version" in tokens:
            # net show version

            output = "NCLU_VERSION={0}\n".format(NCLU_VERSION)

            try:
                with open("/etc/lsb-release") as f:
                    output += f.read()
                command_ok = True
            except OSError as e:
                output += str(e)
                command_ok = False

        elif "pluggables" in tokens:
            # net show interface pluggables [json]
            output = show_pluggables(cli)
            command_ok = True

        elif "vrf" in tokens:
            # net show vrf

            try:
                output = check_output(["/usr/bin/vrf", "list"])
                command_ok = True
            except CalledProcessError as e:
                output = str(e.output)
                command_ok = False

        elif "bfd" in tokens:
            command = ["/usr/bin/ptmctl", "-b"]
            if "detail" in tokens:
                command.append("-d")
            if "json" in tokens:
                command.append("-j")
            command_ok, output = nclu_check_output(command, "ptmd")

        else:
            # The remaining commands are executed with netshow2.

            if "interface" in tokens:
                if "bondmems" in tokens:
                    # net show interface bondmems [json|mac]
                    output = netshow2.net_show_interface_bondmems("mac" in tokens, "json" in tokens)
                elif "bonds" in tokens:
                    # net show interface bonds [json|mac]
                    output = netshow2.net_show_interface_bonds("mac" in tokens, "json" in tokens)
                else:
                    # net show interface all [json|mac]
                    # net show interface [<interface>] [json]
                    # net show interface <interface> detail
                    output = netshow2.net_show_interface(tokens.get("<interface>"), "mac" in tokens, "json" in tokens,
                                                         "detail" in tokens, "all" in tokens)

            elif "system" in tokens:
                # net show system
                output = netshow2.net_show_system()

            elif "lldp" in tokens:
                # net show lldp [<interface>] [json]
                output = netshow2.net_show_lldp(tokens.get("<interface>"), "json" in tokens)

            elif "counters" in tokens:
                # net show counters [json]
                output = netshow2.net_show_counters("json" in tokens)

            else:
                # This is a programming error.
                raise RuntimeError("NCLU couldn't decide how to process the command.")

            command_ok = "error:" not in output.lower()
            if not command_ok:
                # TODO - Modify netshow2.py to avoid having to strip "ERROR" from the output.
                # Have the functions return a (bool, str) 2-tuple.
                output = re.sub(r"""^\s*ERROR:\s*""", '', output, re.IGNORECASE)

        if command_ok is None:
            # This is a programming error.
            raise RuntimeError("NCLU couldn't decide how to process the command.")

        # TODO - Don't print output from wrappers.
        print output
        return command_ok


def get_dhcp_configuration():
    dhcp_relay_servers = []
    dhcp_relay_interfaces = []
    dhcp_servers = []

    dhcp_relay6_servers = []
    dhcp_relay6_interfaces = []
    dhcp_servers6 = []

    relay_key_value_pair_re = re.compile(r'''^\s*(SERVERS|INTF_CMD)\s*=\s*"([^"]*)"''', re.IGNORECASE | re.MULTILINE)
    server_key_value_pair_re = re.compile(r'''^\s*INTERFACES\s*=\s*"([^"]*)"''', re.IGNORECASE | re.MULTILINE)

    # IPv4 and IPv6 relay
    for dhcp_file in [ISC_DHCP_RELAY4, ISC_DHCP_RELAY6]:
        if not isfile(dhcp_file):
            continue

        with open(dhcp_file) as f:
            config_text = f.read()

        key_value_pairs = relay_key_value_pair_re.findall(config_text)

        for key, value in key_value_pairs:
            if key.upper() == "SERVERS":
                if dhcp_file.endswith('6'):
                    # relay6 server
                    interfaces = value.replace("-u", '').split()

                    try:
                        interfaces_as_glob = ifnames_to_glob(interfaces)
                    except (RuntimeError, ValueError) as e:
                        # Could not globify the interfaces.
                        log.warning("(get_dhcp_configuration) \"{0}\" raised an exception in ifnames_to_glob: {1}.".format(str(interfaces), e))
                        interfaces_as_glob = None

                    if interfaces_as_glob:
                        dhcp_relay6_servers.append(interfaces_as_glob)
                    else:
                        dhcp_relay6_servers.extend(interfaces)
                else:
                    # relay server
                    dhcp_relay_servers.extend(value.split())

            elif key.upper() == "INTF_CMD":
                if dhcp_file.endswith('6'):
                    # relay6 interface
                    interfaces = value.replace("-l", '').strip().split()
                else:
                    # relay interface
                    interfaces = value.replace("-i", '').strip().split()

                try:
                    interfaces_as_glob = ifnames_to_glob(interfaces)
                except (RuntimeError, ValueError) as e:
                    # Could not globify the interfaces.
                    log.warning("(get_dhcp_configuration) \"{0}\" raised an exception in ifnames_to_glob: {1}.".format(str(interfaces), e))
                    interfaces_as_glob = None

                if dhcp_file.endswith('6'):
                    if interfaces_as_glob:
                        dhcp_relay6_interfaces.append(interfaces_as_glob)
                    else:
                        dhcp_relay6_interfaces.extend(interfaces)
                else:
                    if interfaces_as_glob:
                        dhcp_relay_interfaces.append(interfaces_as_glob)
                    else:
                        dhcp_relay_interfaces.extend(interfaces)

    # IPv4 and IPv6 server
    for dhcp_file in (ISC_DHCP_SERVER4, ISC_DHCP_SERVER6):
        if not isfile(dhcp_file):
            continue

        with open(dhcp_file) as f:
            config_text = f.read()

        servers = server_key_value_pair_re.findall(config_text)

        for value in servers:
            if dhcp_file.endswith('6'):
                # server6
                dhcp_servers6.extend(value.split())
            else:
                # server
                dhcp_servers.extend(value.split())

    if dhcp_relay_servers:
        dhcp_relay_servers = sort_for_humans(dhcp_relay_servers)

    if dhcp_relay_interfaces:
        dhcp_relay_interfaces = sort_for_humans(dhcp_relay_interfaces)

    if dhcp_servers:
        dhcp_servers = sort_for_humans(dhcp_servers)

    if dhcp_relay6_servers:
        dhcp_relay6_servers = sort_for_humans(dhcp_relay6_servers)

    if dhcp_relay6_interfaces:
        dhcp_relay6_interfaces = sort_for_humans(dhcp_relay6_interfaces)

    if dhcp_servers6:
        dhcp_servers6 = sort_for_humans(dhcp_servers6)

    return (dhcp_relay_servers, dhcp_relay_interfaces, dhcp_servers, dhcp_relay6_servers, dhcp_relay6_interfaces, dhcp_servers6)


def get_configured_hostname():
    """
    Return the hostname as a string, or return None.
    """

    if isfile(ETC_HOSTNAME):
        with open(ETC_HOSTNAME) as f:
            for line in f:
                line = line.strip()

                if line:
                    return line
    return None


def revert_dhcp(is_ipv4, is_relay):
    """
    Undo changes to ISC_DHCP_RELAY4, ISC_DHCP_RELAY6, ISC_DHCP_SERVER4, or ISC_DHCP_SERVER6.
    This is a "do something maybe" function by design because there might not be any
    changes to undo.

    is_ipv4 - A boolean; if True, changes are for IPv4; otherwise, IPv6
    is_relay - A boolean; if True, changes are for relays; otherwise, servers
    """

    if is_ipv4:
        if is_relay:
            # IPv4 relay
            backup = ISC_DHCP_RELAY_SCRATCHPAD4_BACKUP
            configuration = ISC_DHCP_RELAY4
            service = "dhcrelay.service"
        else:
            # IPv4 server
            backup = ISC_DHCP_SERVER_SCRATCHPAD4_BACKUP
            configuration = ISC_DHCP_SERVER4
            service = "dhcpd.service"
    else:
        if is_relay:
            # IPv6 relay
            backup = ISC_DHCP_RELAY_SCRATCHPAD6_BACKUP
            configuration = ISC_DHCP_RELAY6
            service = "dhcrelay6.service"
        else:
            # IPv6 server
            backup = ISC_DHCP_SERVER_SCRATCHPAD6_BACKUP
            configuration = ISC_DHCP_SERVER6
            service = "dhcpd6.service"

    if not isfile(backup):
        # There are no changes to undo.
        return

    shutil.copyfile(backup, configuration)

    try:
        check_output(["/bin/systemctl", "reset-failed", service])
        check_output(["/bin/systemctl", "restart", service])
    except CalledProcessError as e:
        log.error("Failed to reset or restart {0} while reverting {1}.".format(service, configuration))
        log.exception(e)


def revert_hostname():
    if isfile(ETC_HOSTS_SCRATCHPAD_BACKUP):
        shutil.copyfile(ETC_HOSTS_SCRATCHPAD_BACKUP, ETC_HOSTS)

    if isfile(ETC_HOSTNAME_SCRATCHPAD_BACKUP):
        shutil.copyfile(ETC_HOSTNAME_SCRATCHPAD_BACKUP, ETC_HOSTNAME)

    if isfile(ETC_DHCP_EXIT_HOOKS_SCRATCHPAD_BACKUP):
        shutil.copyfile(ETC_DHCP_EXIT_HOOKS_SCRATCHPAD_BACKUP, ETC_DHCP_EXIT_HOOKS)
